import configparser
import os
from pathlib import Path


class Configuration():
    def __init__(self, cfg_dir=os.environ.get('AWSSSO_CONFIG_DIR', '~/.awssso')):
        self._config = configparser.ConfigParser()
        self._config_dir = Path(cfg_dir).expanduser()
        self._config_file = Path(f'{self._config_dir.resolve()}/config')

        self.__ensure_config_dir()
        self.__read()

    def __ensure_config_dir(self):
        self._config_dir.mkdir(exist_ok=True)

    def __read(self):
        self._config.read(self._config_file.resolve())

    @property
    def configfile(self):
        return self._config_file

    @property
    def configdir(self):
        return self._config_dir.resolve()

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, cfg):
        self._config = cfg

    def save(self):
        with self._config_file.open('w') as f:
            self._config.write(f)



import xml.etree.ElementTree as ET
from base64 import b64decode

import boto3
from botocore.exceptions import ClientError


class Error(Exception):
    """Base class for SAMLHelper exceptions."""

    def __init__(self, msg=''):
        self.message = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.message

    __str__ = __repr__


class BotoClientError(Error):
    """Raised when boto call failed."""

    def __init__(self, response):
        Error.__init__(self, response['Error']['Message'])
        self.request_id = response['ResponseMetadata']['RequestId']
        self.args = (response, )


class AssumeRoleValidationError(BotoClientError):
    """Raised when SAML assertion validation failed."""

    pass


class SAMLHelper():
    NS = {
        'a': 'urn:oasis:names:tc:SAML:2.0:assertion'
    }

    XPATH = {
        'roles': ".//a:Assertion/a:AttributeStatement/a:Attribute[@Name='https://aws.amazon.com/SAML/Attributes/Role']/a:AttributeValue",
        'duration': ".//a:Assertion/a:AttributeStatement/a:Attribute[@Name='https://aws.amazon.com/SAML/Attributes/SessionDuration']/a:AttributeValue"
    }

    def __init__(self, encoded_payload):
        # Calling AssumeRoleWithSAML does not require the use of AWS security credentials.
        # The identity of the caller is validated by using keys in the metadata document that is uploaded for the SAML provider entity for your identity provider.
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sts.html#STS.Client.assume_role_with_saml
        self._sts = boto3.client('sts', aws_access_key_id='', aws_secret_access_key='', aws_session_token='')
        self._root = ET.fromstring(b64decode(encoded_payload))
        self._role_arn, self._principal_arn = self._get_roles()
        self._duration = self._get_duration()
        self._payload = encoded_payload

    @property
    def duration(self):
        return self._duration

    def _get_roles(self):
        e = self._root.find(SAMLHelper.XPATH['roles'], SAMLHelper.NS)
        return tuple(e.text.split(','))

    def _get_duration(self):
        e = self._root.find(SAMLHelper.XPATH['duration'], SAMLHelper.NS)
        return int(e.text)

    def assume_role(self, duration=None):
        duration = duration or self._duration
        try:
            return self._sts.assume_role_with_saml(
                RoleArn=self._role_arn,
                PrincipalArn=self._principal_arn,
                SAMLAssertion=self._payload,
                DurationSeconds=duration
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ValidationError':
                raise AssumeRoleValidationError(e.response)
            else:
                raise BotoClientError(e.response)



import pickle
from hashlib import sha256

from selenium import webdriver
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


class Error(Exception):
    """Base class for SSODriver exceptions."""

    def __init__(self, msg=''):
        self.message = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.message

    __str__ = __repr__


class AlertMessage(Error):
    """Raised when alert frame is displayed."""
    pass

class MFACodeNeeded(Error):
    """Raised when MFA code is needed."""

    def __init__(self, mfa_form):
        Error.__init__(self, 'MFA Code needed')
        self.mfa_form = mfa_form
        self.args = (mfa_form, )


class SSODriver():
    def __init__(self, url, username, headless=True, cookie_dir=None):
        self._url = url
        self._cookie_hash = SSODriver.hash(f'{username}@{url}')
        self._cookie_file = f'{cookie_dir}/{self._cookie_hash}.pkl' if cookie_dir else None
        chrome_options = Options()
        if headless:
            chrome_options.add_argument('--headless')
        self._driver = webdriver.Chrome(chrome_options=chrome_options)
        self._driver.implicitly_wait(5)
        self._poll_frequency = 0.05

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    @staticmethod
    def hash(s):
        return sha256(s.encode()).hexdigest()

    def _load_cookies(self):
        try:
            cookies = pickle.load(open(self._cookie_file, 'rb'))
            for cookie in cookies:
                if 'expiry' in cookie:
                    del cookie['expiry']
                self._driver.add_cookie(cookie)
        except FileNotFoundError:
            pass

    def _dump_cookies(self):
        cookies = []
        exclude = ['x-amz-sso_authn']
        for cookie in self._driver.get_cookies():
            if cookie['name'] not in exclude:
                cookies.append(cookie)
        pickle.dump(cookies, open(self._cookie_file, 'wb'))

    def _find_element_by_id(self, element_id, driver=None, timeout=5):
        driver = driver or self._driver
        return WebDriverWait(driver, timeout, poll_frequency=self._poll_frequency).until(
            EC.visibility_of_element_located((By.ID, element_id))
        )

    def _click_element_by_id(self, element_id, driver=None, timeout=5):
        element = self._find_element_by_id(element_id, driver, timeout)
        element.click()
        return element

    def refresh_token(self, username, password, restore=False):
        self.get()
        self.login(username, password)
        try:
            return self.get_token(restore)
        except TimeoutException:
            self.check_alert()
            self.check_mfa()

    def get_token(self, restore=False):
        WebDriverWait(self._driver, 1, poll_frequency=self._poll_frequency).until(
            EC.presence_of_element_located((By.TAG_NAME, 'portal-dashboard'))
        )
        self._driver.get(self._url)
        cookie = self._driver.get_cookie('x-amz-sso_authn')
        if restore:
            self._driver.back()
        return (cookie['value'], cookie['expiry'])

    def close(self):
        if self._cookie_file:
            self._dump_cookies()
        self._driver.quit()

    def get(self):
        self._driver.get(self._url)
        if self._cookie_file:
            self._load_cookies()
        return self._driver

    def login(self, username, password):
        el_username = self._find_element_by_id('wdc_username')
        el_password = self._find_element_by_id('wdc_password')
        el_signin = self._find_element_by_id('wdc_login_button')

        el_username.clear()
        el_username.send_keys(username)
        el_password.clear()
        el_password.send_keys(password)
        el_signin.click()

    def check_alert(self):
        try:
            wait = WebDriverWait(self._driver, 1, poll_frequency=self._poll_frequency)
            alert = wait.until(EC.presence_of_element_located((By.ID, 'alertFrame')))
            error = alert.find_element_by_css_selector('div.a-alert-error > div.a-box-inner > h4').text
            message = alert.find_element_by_css_selector('div.a-alert-error > div.a-box-inner > div.gwt-Label').text
            raise AlertMessage(f'{error}: {message}')
        except (TimeoutException, NoSuchElementException):
            pass

    def check_mfa(self):
        try:
            wait = WebDriverWait(self._driver, 1, poll_frequency=self._poll_frequency)
            mfa = wait.until(EC.presence_of_element_located((By.ID, 'mfa_form')))
            raise MFACodeNeeded(mfa)
        except TimeoutException:
            pass

    def send_mfa(self, mfa_form, mfacode, trusted_device=True):
        el_mfacode = self._find_element_by_id('wdc_mfacode', mfa_form)
        el_mfacheckbox = self._find_element_by_id('wdc_mfacheckbox', mfa_form)
        el_signin = self._find_element_by_id('wdc_login_button', mfa_form)

        el_mfacode.clear()
        el_mfacode.send_keys(mfacode)
        if trusted_device:
            el_mfacheckbox.click()
        el_signin.click()



import requests

class SSOClient():
    def __init__(self, token, region='eu-west-1'):
        self._token = token
        self._region = region
        self._s = requests.Session()
        self._s.headers.update({
            'x-amz-sso_bearer_token': self._token
        })

    def whoami(self):
        r = self._s.get(f'https://portal.sso.{self._region}.amazonaws.com/token/whoAmI')
        return r.json()

    def get_instances(self):
        r = self._s.get(f'https://portal.sso.{self._region}.amazonaws.com/instance/appinstances')
        return [i for i in r.json()['result'] if i['applicationName'] == 'AWS Account']

    def get_profiles(self, instance_id):
        r = self._s.get(f'https://portal.sso.{self._region}.amazonaws.com/instance/appinstance/{instance_id}/profiles')
        return r.json()['result']

    def get_saml_payload(self, instance_id, profile_id):
        for profile in self.get_profiles(instance_id):
            if profile['id'] == profile_id:
                url = profile['url']
        r = self._s.get(url)
        return r.json()['encodedResponse']



import argparse
import os
import sys
import webbrowser
from time import time

import inquirer
from halo import Halo

from awssso import __version__
from awssso.config import Configuration
from awssso.helpers import (SPINNER_MSGS, CredentialsHelper, SecretsManager,
                            config_override, validate_empty, validate_url)
from awssso.saml import AssumeRoleValidationError, BotoClientError, SAMLHelper
from awssso.ssoclient import SSOClient
from awssso.ssodriver import AlertMessage, MFACodeNeeded, SSODriver


def __refresh_token(url, username, password, config_dir, headless=True, spinner=True):
    spinner = Halo(enabled=spinner)
    try:
        spinner.start(SPINNER_MSGS['token_refresh'])
        driver = SSODriver(url, username, headless=headless, cookie_dir=config_dir)
        try:
            return driver.refresh_token(username, password)
        except MFACodeNeeded as e:
            spinner.stop()
            mfacode = inquirer.text(message='MFA Code')
            spinner.start(SPINNER_MSGS['mfa_send'])
            driver.send_mfa(e.mfa_form, mfacode)
            spinner.start(SPINNER_MSGS['token_refresh'])
            return driver.get_token()
        except AlertMessage as e:
            sys.exit(e)
        finally:
            spinner.stop()
    except KeyboardInterrupt as e:
        spinner.stop()
        raise e
    finally:
        driver.close()


def __get_or_refresh_token(url, username, password, secrets, config_dir, force_refresh=False, headless=True, spinner=True):
    token = secrets.get('authn-token')
    stored_password = secrets.get('credentials')
    expiry_date = int(secrets.get('authn-expiry-date', '0'))
    if (force_refresh) or (not token) or (time() > expiry_date) or (stored_password != password):
        token, expiry_date = __refresh_token(url, username, password, config_dir, headless, spinner)
        if stored_password != password:
            secrets.set('credentials', password)
        secrets.set('authn-token', token)
        secrets.set('authn-expiry-date', str(expiry_date))
    return token


def configure(args):
    profile = args.profile
    cfg = Configuration()
    params = config_override(cfg.config, profile, args)

    try:
        inquirer.prompt([
            inquirer.Text('url', message='URL', default=params.get('url', ''), validate=validate_url),
            inquirer.Text('aws_profile', message='AWS CLI profile', default=params.get('aws_profile', profile), validate=validate_empty),
            inquirer.Text('username', message='Username', default=params.get('username', ''), validate=validate_empty)
        ], answers=params, raise_keyboard_interrupt=True)
        secrets = SecretsManager(params.get('username'), params.get('url'))
        password = inquirer.password(message='Password', default=secrets.get('credentials', ''), validate=validate_empty)

        token = __get_or_refresh_token(
            params['url'], params['username'], password,
            secrets, cfg.configdir, args.force_refresh, args.headless, args.spinner
        )
        sso = SSOClient(token, params['region'])

        instances = sso.get_instances()
        inquirer.prompt([
            inquirer.List(
                'instance_id',
                message='AWS Account',
                choices=[(_['name'], _['id']) for _ in instances]
            )
        ], answers=params, raise_keyboard_interrupt=True)

        profiles = sso.get_profiles(params['instance_id'])
        inquirer.prompt([
            inquirer.List(
                'profile_id',
                message='AWS Profile',
                choices=[(_['name'], _['id']) for _ in profiles]
            )
        ], answers=params, raise_keyboard_interrupt=True)

        cfg.save()
    except KeyboardInterrupt:
        sys.exit(1)


def login(args):
    profile = args.profile
    cfg = Configuration()

    if profile not in cfg.config:
        sys.exit(f'profile {profile} does not exist, use "awssso configure -p {profile}" to create it')

    params = config_override(cfg.config, profile, args)
    aws_profile = params.get('aws_profile', profile)
    secrets = SecretsManager(params.get('username'), params.get('url'))
    password = secrets.get('credentials')

    if not password:
        sys.exit(f'Cannot get password from secrets, run "awssso configure -p {profile}"')

    try:
        token = __get_or_refresh_token(
            params['url'], params['username'], password,
            secrets, cfg.configdir, args.force_refresh, args.headless, args.spinner
        )
        sso = SSOClient(token, params['region'])

        if args.interactive:
            instances = sso.get_instances()
            inquirer.prompt([
                inquirer.List(
                    'instance_id',
                    message='AWS Account',
                    choices=[(_['name'], _['id']) for _ in instances]
                )
            ], answers=params, raise_keyboard_interrupt=True)

            profiles = sso.get_profiles(params['instance_id'])
            inquirer.prompt([
                inquirer.List(
                    'profile_id',
                    message='AWS Profile',
                    choices=[(_['name'], _['id']) for _ in profiles]
                )
            ], answers=params, raise_keyboard_interrupt=True)

        payload = sso.get_saml_payload(params['instance_id'], params['profile_id'])
        saml = SAMLHelper(payload)
        credentials = saml.assume_role(args.duration)['Credentials']

        ch = CredentialsHelper(credentials)

        if args.export:
            print(ch.configure_export())
        elif args.json:
            print(ch.configure_json())
        elif args.console:
            session_duration = args.duration or saml.duration
            signin_url = ch.console_signin(session_duration)
            print(signin_url)
            if args.browser:
                webbrowser.open_new_tab(signin_url)
        else:
            ch.configure_cli(aws_profile)
    except (AssumeRoleValidationError, BotoClientError) as e:
        sys.exit(f'{e} (request id: {e.request_id})')
    except KeyboardInterrupt:
        sys.exit(1)


class DurationAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values < 900:
            parser.error(f'argument {option_string}: minimum value is 900')
        setattr(namespace, self.dest, values)


def main():
    default_region = os.environ.get('AWSSSO_REGION')
    default_region = default_region or os.environ.get('AWS_DEFAULT_REGION', 'eu-west-1')

    default_profile = 'default'
    default_aws_profile = os.environ.get('AWS_PROFILE')
    default_aws_profile = default_aws_profile or os.environ.get('AWS_DEFAULT_PROFILE')

    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--region', default=default_region, help='AWS SSO region (default: AWSSSO_REGION, AWS_DEFAULT_REGION, eu-west-1)')
    parser.add_argument('--no-headless', dest='headless', action='store_false', default=True, help='show web browser')
    parser.add_argument('--no-spinner', dest='spinner', action='store_false', default=True, help='disable all spinners')
    subparsers = parser.add_subparsers(title='subcommands')

    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('-p', '--profile', default=default_profile, help=f'AWS SSO Profile (default: {default_profile})')
    parent_parser.add_argument('-a', '--aws-profile', default=default_aws_profile, help='AWS CLI Profile (default: AWS_PROFILE, fallback: same as --profile)')
    parent_parser.add_argument('-f', '--force-refresh', action='store_true', default=False, help='force token refresh')

    configure_parser = subparsers.add_parser('configure', parents=[parent_parser])
    configure_parser.add_argument('--url')
    configure_parser.add_argument('--username')
    configure_parser.set_defaults(func=configure)

    login_parser = subparsers.add_parser('login', parents=[parent_parser])
    login_parser.add_argument('-d', '--duration', action=DurationAction, type=int, help='duration (seconds) of the role session (default/maximum: from SAML payload, minimum: 900)')
    login_parser_group = login_parser.add_mutually_exclusive_group()
    login_parser_group.add_argument('-e', '--export', action='store_true', default=False, help='output credentials as environment variables')
    login_parser_group.add_argument('-j', '--json', action='store_true', default=False, help='output credentials in JSON format (see https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sourcing-external.html)')
    login_parser_group.add_argument('-c', '--console', action='store_true', default=False, help='output AWS Console Sign In url')
    login_parser.add_argument('-b', '--browser', action='store_true', default=False, help='open web browser with AWS Console Sign In url')
    login_parser.add_argument('-i', '--interactive', action='store_true', default=False, help='interactively choose AWS account and role')
    login_parser.set_defaults(func=login)

    args = parser.parse_args()

    try:
        func = args.func
        if callable(func):
            func(args)
    except AttributeError:
        parser.print_help(sys.stderr)



#!/usr/bin/env python
from setuptools import find_packages, setup

requires = [
    'awscli>=1.16.10,<2.0.0',
    'boto3>=1.9.0,<2.0.0',
    'halo',
    'inquirer>=2.6.0,<3.0.0',
    'keyring>=19.0.0,<20.0.0',
    'requests',
    'secretstorage; platform_system == "Linux"',
    'selenium>=3.14.0,<4.0.0',
]

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup_options = dict(
    name='awssso',
    use_scm_version=True,
    description='Command Line tool for AWS SSO Credentials',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='wnkz',
    author_email='wnkz@users.noreply.github.com',
    url='http://github.com/wnkz/aws-sso',
    entry_points={
        'console_scripts': [
            'awssso=awssso.cli:main'
        ],
    },
    packages=find_packages(exclude=['tests*']),
    install_requires=requires,
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'Natural Language :: English',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: MacOS',
        'Operating System :: POSIX',

        'Programming Language :: Python :: 3.7'
    ],
    keywords=['aws', 'sso', 'cloud', 'cli', 'credentials']
)

setup(**setup_options)



import json
import re
import subprocess
from urllib.parse import urlparse

import keyring
import requests
from inquirer.errors import ValidationError

SPINNER_MSGS = {
    'token_refresh': 'Refreshing token',
    'mfa_send': 'Sending MFA code'
}


class SecretsManager():
    def __init__(self, username, url):
        self._username = username
        self._base_service_name = f'awssso.{urlparse(url).netloc}'

    def get(self, stype, default=None):
        return keyring.get_password(
            f'{self._base_service_name}.{stype}',
            self._username
        ) or default

    def set(self, stype, password):
        return keyring.set_password(
            f'{self._base_service_name}.{stype}',
            self._username,
            password
        )


class CredentialsHelper():
    CNAMES = {
        'AccessKeyId': {
            'awscli': 'aws_access_key_id',
            'env': 'AWS_ACCESS_KEY_ID',
            'console': 'sessionId'
        },
        'SecretAccessKey': {
            'awscli': 'aws_secret_access_key',
            'env': 'AWS_SECRET_ACCESS_KEY',
            'console': 'sessionKey'
        },
        'SessionToken': {
            'awscli': 'aws_session_token',
            'env': 'AWS_SESSION_TOKEN',
            'console': 'sessionToken'
        }
    }

    def __init__(self, credentials):
        self._credentials = credentials

    def configure_cli(self, profile):
        for _ in self._credentials:
            if _ in CredentialsHelper.CNAMES:
                subprocess.run([
                    'aws', 'configure',
                    '--profile', profile,
                    'set', CredentialsHelper.CNAMES[_]['awscli'], self._credentials[_]
                ])

    def configure_export(self, profile=None):
        exports = []
        for _ in self._credentials:
            if _ in CredentialsHelper.CNAMES:
                export = f'export {CredentialsHelper.CNAMES[_]["env"]}={self._credentials[_]}'
                exports.append(export)
        return '\n'.join(exports)

    def configure_json(self):
        export = {
            **self._credentials,
            **{
                'Version': 1,
                'Expiration': self._credentials['Expiration'].isoformat()
            }
        }
        return json.dumps(export)

    def console_signin(self, duration):
        session = {}
        for _ in self._credentials:
            if _ in CredentialsHelper.CNAMES:
                session.update({
                    CredentialsHelper.CNAMES[_]['console']: self._credentials[_]
                })

        params = {
            'Action': 'getSigninToken',
            'Session': json.dumps(session),
            'SessionDuration': duration
        }
        response = requests.get('https://signin.aws.amazon.com/federation', params=params)

        login_request = requests.Request(
            'GET',
            'https://signin.aws.amazon.com/federation',
            params={
                'Action': 'login',
                'Destination': 'https://console.aws.amazon.com/',
                'SigninToken': response.json()['SigninToken']
            }
        ).prepare()

        return login_request.url


def config_override(config, section, args, keep=['url', 'region', 'username', 'aws_profile']):
    if section not in config:
        config[section] = {}
    params = config[section]

    for arg in vars(args):
        value = getattr(args, arg)
        if (not keep) or (arg in keep and value is not None):
            params[arg] = value
    return params


def validate_url(answers, url):
    rx = r'^https://[\S-]+\.awsapps\.com/start/$'
    if re.match(rx, url) is None:
        raise ValidationError('', reason=f'URL must match {rx}')
    return True


def validate_empty(answers, s):
    if not s:
        raise ValidationError('', reason='Must not be empty')
    return True
